class RMSWarning(Warning):
    pass
class KernelError(Exception):
    pass
class PSFNormalizationWarning(Warning):
    pass
class MaskWarning(Warning):
    pass

class ShapeMatchError(Exception):
    pass